package server

import (
	"context"
	"errors"
	"log"
	"math/rand"
	"sync"
	"time"
)

// topic represents a channel to which subscribers can subscribe, and publishers
// can publish a message
type topic struct {
	id          string
	subscribers map[int]subscriber
	messages    int
	last        time.Time
	ctx         context.Context
	cancel      context.CancelFunc
	mu          sync.Mutex
}

// subscriber is a function that is called for every new message on a topic
type subscriber func(msg *message) error

// newTopic creates a new topic
func newTopic(id string) *topic {
	ctx, cancel := context.WithCancel(context.Background())
	return &topic{
		id:          id,
		subscribers: make(map[int]subscriber),
		last:        time.Now(),
		ctx:         ctx,
		cancel:      cancel,
	}
}

func (t *topic) Subscribe(s subscriber) int {
	t.mu.Lock()
	defer t.mu.Unlock()
	subscriberID := rand.Int()
	t.subscribers[subscriberID] = s
	t.last = time.Now()
	return subscriberID
}

func (t *topic) Unsubscribe(id int) int {
	t.mu.Lock()
	defer t.mu.Unlock()
	delete(t.subscribers, id)
	return len(t.subscribers)
}

func (t *topic) Publish(m *message) error {
	t.mu.Lock()
	defer t.mu.Unlock()
	if len(t.subscribers) == 0 {
		return errors.New("no subscribers")
	}
	t.last = time.Now()
	t.messages++
	for _, s := range t.subscribers {
		if err := s(m); err != nil {
			log.Printf("error publishing message to subscriber")
		}
	}
	return nil
}

func (t *topic) Stats() (subscribers int, messages int) {
	t.mu.Lock()
	defer t.mu.Unlock()
	return len(t.subscribers), t.messages
}

func (t *topic) Close() {
	t.cancel()
}
